/*

distance=sqrt (x2-x1)^2 + (y2-y1)^2 + (z2-z1)^2
normalize=v2/theLengthOfv2
*/

struct penMath
(
	/******************************
		Name: Return local rotation from object
		Axis Orders:
			1 - XYZ
			2 - XZY
			3 - YZX
			4 - YXZ
			5 - ZXY
			6 - ZYX
			7 - XYX
			8 - YZY
			9 - ZXZ
	*******************************/
	fn localRot
		c:undefined	--Object to extract rotation from.
		p:undefined	--Parent object or any other object to find the local rotation offset from.
		o:1 		--Axis order to calculate the Euler angles with.
		=
	(
		if c!=undefined then 
		(
			if p==undefined then p=c.parent
			quatToEuler (c.transform * (inverse p.transform)).rotation order:o
		)else
		(
			eulerAngles 0 0 0
		)
	),
	
	/******************************
		Name: Hermite Function
		Set the input value from 0 to 1
		Adjust the t1 and t2 value to adjust exceleration curve.
		Defaults are set to create an ease out ease in curve. 
		t1 nd t2 values to get the same ease in and out curve. 
		*Use in conjuction with rangeClamp if your input value is not between 0 and 1.
	*******************************/
	fn hermite
		t1:3.0	--Tangent 1
		t2:2.0	--Tangent 2
		v:0.0	--Input value from 0 to 1
	=
	(
		h1 = -2*v^t1 + 3*v^t2 --Hermite Fn
	),
	
	/******************************
	Cubic Hermite
	
	The default tangent values, m1 and m2 produce a typical ease out, ease in curve. 
	This doesn't work the same way that a bezier function works with the tangents. 
	*******************************/
	fn cubicHermite
		p1:[0,0,0]
		p2:[1,0,1]
		m1:[1,0,0]
		m2:[1,0,0]
		t:0.0
	=
	(
		p=	(2*t^3-3*t^2+1)*p1+
			(t^3-2*t^2+t)*m1+
			(-2*t^3+3*t^2)*p2+
			(t^3-t^2)*m2
	),
	
	/*****************************
		Name:Clamp Range
		Author: Paul Neale
		Client: PEN Productions Inc.
		Last Updated: May 30 2005
		Version: 0.02.0
		
		ToDo:
		
		Bugs:
		
		Usage:
		Example:1 
		clampRange stVal:0.0 edVal:50.0 nStVal:0.0 nEdVal:100.0 val:25.0 clamp:false printVals:true
		Example:2 Invert a range
		clampRange stVal:-1.0 edVal:1.0 nStVal:1.0 nEdVal:-1.0 val:1.0 clamp:true printVals:true
	******************************/
	fn clampRange 
	stVal:0.0  			--Start value of in comming range.
	edVal:100.0 		--End value of incomming range.
	nStVal:0.0 			--Start value of out going range.
	nEdVal:1.0 			--End value of out going range.
	val:0.0 			--In value that will be converted.
	clamp:false			--If true will return a value that is clamped in the out going range. 
	=
	(
		local outVal
		
		--Get the percent of the inVal
		inRange=(edVal as float)-(stVal as float)
		inVal=(val as float)-(stVal as float)
		valPercent=inVal/inRange
		
		--Get the value of the based on the new range.
		nRange=(nEdVal as float)-(nStVal as float)
		
		--Invert the math if the nStVal> nEdVal so that values ranges can be inverted.
		if nStVal<=nEdVal then
		(
			outVal=(nRange*valPercent)-(abs nStVal)
		)else
		(
			outVal=(nRange*valPercent)+(abs nStVal)
		)
		
		--Clamp out going value to out range if clamp is true.
		if clamp then
		(
			if outVal<nStVal then (return nStVal)
			if outVal>nEdval then (return nEdVal)
		)
		return outVal
	),
	
	/*****************************
		Name:			Mirror Matrix w/ No Scale
		Author: 		Paul Neale
		Client: 		PEN Productions Inc.
		Last Updated: 	June 10 2005
		Version: 		0.1.0
		
		Arguments:
		axis: Axis as string to mirror over. Options:("x" "y" "z")
		flip: Axis to flip. Options:("x" "y" "z")
		tm: Transform matrix to mirror.
		pivotTm: Pivot point matrix to mirror around. Defaults to the identity matrix.
		
		Usage:
		Create 3 point helpers and run these three lines of code.
		tm=$point01.transform
		pivotTm=$point02.transform
		$point03.transform=penMath.mirrorMatrix axis:"x" flip:"y" tm:tm pivotTm:pivotTm
	******************************/
	fn mirrorMatrixOld
		axis:"x"			--Axis to mirror over
		flip:"x"			--Axis to flip
		tm:(matrix3 1)		--Matrix to mirror
		pivotTm:(matrix3 1)	--Matrix to mirror around
		=
	(
		--Get the local tm of the input matrix.
		local localTm=tm*inverse pivotTm

		--Set the multiplier for the mirror axis.
		
		
		mAxis=case axis of
		(
			"x": [-1,1,1]
			"y": [1,-1,1]
			"z": [1,1,-1]
		)
			
		--Mirror the matrix vectors.
		vPos=case flip of
		(
			"x": 
			(
				y=localTm[2]*mAxis	--Mirror Y
				z=localTm[3]*mAxis	--Mirror Z
				x=cross y z			--Flip X
				#(x,y,z)
			)
			"y":
			(
				x=localTm[1]*mAxis	--Mirror X
				z=localTm[3]*mAxis	--Mirror Z
				y=cross z x			--Flip Y
				#(x,y,z)
			)
			"z":
			(
				x=localTm[1]*mAxis	--Mirror X
				y=localTm[2]*mAxis	--Mirror Y
				z=cross x y			--Flip Z
				#(x,y,z)
			)
		)
		
		fAxis=case flip of
		(
			"x": [-1,1,1]
			"y": [1,-1,1]
			"z": [1,1,-1]
		)
		mReflect = scalematrix mAxis
		fReflect = scalematrix fAxis
				
		--Set the matrix in world space
		mTm=(matrix3 vPos[1] vPos[2] vPos[3] (localTm.row4*mAXis))*pivotTm	--Convert to world space.

		mtm2 = xformMat pivotTm (mReflect * (xformMat tm fReflect)) 
		format "check:\nmtm: %\nmtm2:%\n" mTm mTm2

		mTm																	--Return the mirrored matrix.		
	
	),

	/***************************
	* Mirror Matrix
	* Allows for non-uniform scale, preserves 'handedness' of tm
	***************************/
	fn mirrorMatrix
		axis:"x"			--Axis to mirror over
		flip:"x"			--Axis to flip
		tm:(matrix3 1)		--Matrix to mirror
		pivotTm:(matrix3 1)	--Matrix to mirror around
		=
	(
		fn FetchReflection a = 
		(
			case a of (
			"x": [-1,1,1]  -- reflect in YZ plane
			"y": [1,-1,1]  --         in ZX plane
			"z": [1,1,-1]  --         in XY plane
			)
		)
		
		aReflection = scalematrix (FetchReflection axis)		
		fReflection = scalematrix (FetchReflection flip)
		
		-- First reflection mirrors the transform in 'pivot' space, but inverts orientation and
		-- leaves us with a negative scale.
		-- ...so we need do a second reflection (in the 'tm' space) to restore the orientation
			
		--Return the mirrored matrix	
		fReflection * (tm * (inverse pivotTm)) * aReflection * pivotTm 
	),
	
	/***************************
	Returns if a matrix is right or left handed. 
	Arguments:
	Takes one argument in the form of a matrix3 value. 
	***************************/
	fn getHandedMatrix tm=
	(
		x=normalize tm[1] 
		y=normalize tm[2] 
		z=normalize tm[3] 
		an=acos (dot (cross x y ) z)
		hand=if an<=90 then #right else #left
	),
	
	/***************************
	Projects a point in front of B on a triangle that is perpendicular to b
	***************************/
	fn projectSwivelAngle A B C multi:2=
	(
		A=if (isKindOf A node) then A.pos else A
		B=if (isKindOf B node) then B.pos else B
		C=if (isKindOf C node) then C.pos else C

		aa=distance B C
		bb=distance A C
		cc=distance A B
		
		multiMid=cc/(cc+aa)
		bbMidPos=A+((normalize (C-A)*(bb*multiMid)))
		L=length (B-bbMidPos)
		dir=normalize (B-bbMidPos)
		bbMidPos+=dir*L+(dir*multi)
	),
	
	/***************************
	Cubic Bezier N function
	Works on n vectors and uses the number of vectors passed to it to determine 
		the number of iterations. 
	
	p: an array of point3 values. 
	t: time or percent along the path from 0 to 1. 
	
	Usage:
	$test.pos=bezierN #($point01.pos,$point02.pos) .5
	$test.pos=bezierN #($point01.pos,$point02.pos,$point03.pos) .2
	***************************/
	fn bezierN p t=
	(
		n=p.count --Number of iterations
		while (n>1) do --Loop until only one vector is left
		(
			for i = 1 to (n-1) do --Loop through vectors and calc mid point. 
			(
				p[i]=((1-t)*p[i])+(t*p[i+1]) --Calc the mid point between each vector
			)
			n-=1 --Reduce the count for each pass until one vector is left. 
		)
		p[1] --Return the vector. 
	),
	
	/************************************************
	Does an == comparison between the two numbers
	Takes into account max's rounding issues.
	************************************************/
	fn closeEnough val compareVal variance:0.00001=
	(
		if (classOf val)==(classOf compareVal) then
		(
			valAr=#()
			compAr=#()
			case (classOf val) of
			(
				float: (append valAr val; append compAr compareVal)
				point2: for i = 1 to 2 do (append valAr val[i]; append compAr compareVal[i])
				point3: for i = 1 to 3 do (append valAr val[i];append compAr compareVal[i])
				eulerAngles: 
				(
					append valAr val.x;append valAr val.y;append valAr val.z
					append compAr compareVal.x;append compAr compareVal.y;append compAr compareVal.z
				)
			)
			
			for i = 1 to valAr.count do
			(
				if ((valAr[i]<=compAr[i]+variance)and(valAr[i]>=compAr[i]-variance))==false then (return false)--else (return true)
			)
		)
		true
	),
	
	/************************************************
	Average weighted Vectors
	Used for constraining one vector to several others that are weighted.
	************************************************/
	fn averageWeightedVector vectorArray:#() weightArray:#()=
	(
		v=[0,0,0]
		if vectorArray.count==weightArray.count then
		(
			w=0
			for i = 1 to vectorArray.count do
			(
				w+=weightArray[i]
				v+=vectorArray[i]*weightArray[i]
			)
			v/w
		)else
		(
			format "Error: VectorArray and weightArray must have the same number of elements!\n"
			v
		)
	),
	
	/************************************************
	Returns a weighted average transform matrix
	************************************************/
	fn averageWeightedMatrix matrixArray:#() weightArray:#()=
	(
		aM=matrix3 1
		if matrixArray.count==weightArray.count then
		(
			for vi = 1 to 4 do
			(
				vArray=for i = 1 to matrixArray.count collect matrixArray[i][vi]
				aM[vi]=(averageWeightedVector vectorArray:vArray weightArray:weightArray)
				if vi<4 then aM[vi]=normalize aM[vi]
			)
			orthogonalize aM
		)else
		(
			format "Error: matrixArray and weightArray must have the same number of elements!\n"
			aM
		)
	)
	
)
penMath=penMath()

/***test Area***
aV=penMath.averageWeightedVector vectorArray:#($b1.pos,$b2.pos,$b3.pos) weightArray:#(1,1,1)
$s.pos=aV
aM=penMath.averageWeightedMatrix matrixArray:#($b1.transform,$b2.transform,$b3.transform) weightArray:#(1,.5,1)
$s.transform=aM


penMath.hermite()

--test for mirrorMatrix fn
tm=$point01.transform
pivotTm=$point02.transform
$point03.transform=penMath.mirrorMatrix axis:"x" flip:"x" tm:$point01.transform pivotTm:$point02.transform
$point03.transform=penMath.mirrorMatrix2 axis:"x" flip:"x" tm:$point01.transform pivotTm:$point02.transform

$point03.transform=penMath.mirrorMatrix2 axis:"y" flip:"y" tm:$point01.transform pivotTm:$point02.transform

text01 = text pos:[-200,100,0] text:"text01"
text02 = text pos:[-200,  0,0] text:"text02"
pivPnt = point pos:[10, -10,0] name:"pivPnt"

text02.transform=penMath.mirrorMatrix axis:"x" flip:"x" tm:text01.transform pivotTm:pivPnt.transform
text02.transform=penMath.mirrorMatrix2 axis:"x" flip:"x" tm:text01.transform pivotTm:pivPnt.transform

-- give text01 a left-handed matrix 
text01.transform = scalematrix [-1,1,1] * transmatrix text01.transform.translation 
text02.transform=penMath.mirrorMatrix axis:"x" flip:"x" tm:text01.transform pivotTm:pivPnt.transform
text02.transform=penMath.mirrorMatrix2 axis:"x" flip:"x" tm:text01.transform pivotTm:pivPnt.transform

-- give text01 a non-uniform scale
text01.transform =  scalematrix [-1,2,3] * transmatrix text01.transform.translation
text02.transform=penMath.mirrorMatrix axis:"x" flip:"x" tm:text01.transform pivotTm:pivPnt.transform
text02.transform=penMath.mirrorMatrix2 axis:"x" flip:"x" tm:text01.transform pivotTm:pivPnt.transform

-- reset text01 and give pivot point an ugly scale
text01.transform = transmatrix text01.transform.translation
pivPnt.transform = (rotateZmatrix 35) * scalematrix [-2,3,1] * transmatrix pivPnt.transform.translation
text02.transform=penMath.mirrorMatrix axis:"x" flip:"x" tm:text01.transform pivotTm:pivPnt.transform
text02.transform=penMath.mirrorMatrix2 axis:"x" flip:"x" tm:text01.transform pivotTm:pivPnt.transform

***************************************************
From CG Talk:Posted by Georg Duemlein

struct rpEFn (
 	-- ---
 	-- Easing Equotations by Robert Penner
 	-- http://www.robertpenner.com/
 	-- The BSD License: http://www.opensource.org/licenses/bsd-license.php
 	-- ---
 	-- maxScript port by Georg Duemlein
 	-- 2007-04-18
 	-- http://www.preset.de
 	-- ---
 	-- previews: http://www.gizma.com/easing/
 	-- ---
 	-- t: current time
 	-- b: start value
 	-- c: end value
 	-- d: duration
 	-- ---
 	-- simple linear tweening - no easing, no acceleration
 	fn linearTween t b c d = (
 			c*t/d + b
 	),
 	--  quadratic easing in - accelerating from zero velocity
 	fn easeInQuad t b c d = (
 		t /= d
 		c*t*t + b
 	),
 	-- quadratic easing out - decelerating to zero velocity
 	fn easeOutQuad t b c d = (
 		t /= d
 		-c * t * (t-2) + b
 	),
 	-- quadratic easing in/out - acceleration until halfway, then deceleration
 	fn easeInOutQuad t b c d = (
 		t /= d/2
 		if (t < 1) then (
 			c / 2 * t * t + b
 		)else(
 			t-=1
 			-c / 2 * (t* (t - 2) - 1) + b
 		)
 	),
 	-- cubic easing in - accelerating from zero velocity
 	fn easeInCubic t b c d = (
 		t /= d
 		c*t*t*t + b
 	),
 	-- cubic easing out - decelerating to zero velocity
 	fn easeOutCubic t b c d = (
 		t /= d
 		t-=1
 		c*(t*t*t + 1) + b
 	),
 	-- cubic easing in/out - acceleration until halfway, then deceleration
 	fn easeInOutCubic t b c d = (
 		t /= d/2
 		if (t < 1) then (
 			c/2*t*t*t + b
 		)else(
 			t -= 2
 			c/2*(t*t*t + 2) + b
 		)
 	),
 	-- quartic easing in - accelerating from zero velocity
 	fn easeInQuart t b c d = (
 		t /= d
 		c*t*t*t*t + b
 	),
 	-- quartic easing out - decelerating to zero velocity
 	fn easeOutQuart t b c d = (
 		t /= d
 		t -= 1
 		-c * (t*t*t*t - 1) + b
 	),
 	-- quartic easing in/out - acceleration until halfway, then deceleration
 	fn easeInOutQuart t b c d = (
 		t /= d/2
 		if (t < 1) then(
 			c/2*t*t*t*t + b
 		)else(
 			t -= 2
 			-c/2 * (t*t*t*t - 2) + b
 		)
 	),
 	-- quintic easing in - accelerating from zero velocity
 	fn easeInQuint t b c d = (
 		t /= d
 		c*t*t*t*t*t + b
 	),
 	-- quintic easing out - decelerating to zero velocity
 	fn easeOutQuint t b c d = (
 		t /= d
 		t -= 1
 		c*(t*t*t*t*t + 1) + b
 	),
 	-- quintic easing in/out - acceleration until halfway, then deceleration
 	fn easeInOutQuint t b c d = (
 		t /= d/2
 		if (t < 1) then (
 			c/2*t*t*t*t*t + b
 		)else(
 			t -= 2
 			c/2*(t*t*t*t*t + 2) + b
 		)
 	),
 	-- sinusoidal easing in - accelerating from zero velocity
 	fn easeInSine t b c d = (
 		-c * cos (t/d * (180 / 2)) + c + b
 	),
 	-- sinusoidal easing out - decelerating to zero velocity
 	fn easeOutSine t b c d = (
 		c * sin (t/d * 90) + b
 	),
 	-- sinusoidal easing in/out - accelerating until halfway, then decelerating
 	fn easeInOutSine t b c d = (
 		-c/2 * (cos(180 * t/d) - 1) + b
 	),
 	-- exponential easing in - accelerating from zero velocity
 	fn easeInExpo t b c d = (
 		c * pow 2 (10 * (t/d - 1)) + b
 	),
 	-- exponential easing out - decelerating to zero velocity
 	fn easeOutExpo t b c d = (
 		c *  (-pow 2 (-10 * t/d) + 1 ) + b
 	),
 	-- exponential easing in/out - accelerating until halfway, then decelerating
 	fn easeInOutExpo t b c d = (
 		t /= d/2
 		if (t < 1) then (
 			c / 2 * pow 2 (10. * (t - 1)) + b
 		)else(
 			t -= 1
 			c / 2 *  (-pow 2 (-10. * t) + 2 ) + b
 		)
 	),
 	-- circular easing in - accelerating from zero velocity
 	fn easeInCirc t b c d = (
 		t /= d
 		-c * (sqrt(1 - t*t) - 1) + b
 	),
 	-- circular easing out - decelerating to zero velocity
 	fn easeOutCirc t b c d = (
 		t /= d
 		t -= 1
 		c * sqrt(1 - t*t) + b
 	),
 	-- circular easing in/out - acceleration until halfway, then deceleration
 	fn easeInOutCirc t b c d = (
 		t /= d/2
 		if (t < 1) then (
 			-c/2. * (sqrt(1. - t*t) - 1.) + b
 		)else(
 			t -= 2
 			c/2. * (sqrt(1. - t*t) + 1.) + b
 		)
 	)
 )
 
 -- sample
 d = 50.
 t = 0.
 	
 while t < d do (
 	v1 = v2 = v3 = 0
 	v1 = rpEFn.linearTween t 0. 100. d
 	v2 = rpEFn.easeInOutQuad t 1. 100. d
 	v3 = rpEFn.easeInOutQuad t 0. 1. (d/2)
 	sp = sphere pos:[v1, 0, v2] radius:(.5 + v3)
 	sp.wirecolor = [255,255,255] * v3
 	-- ---
 	t += .5
 	max zoomext sel all
 )
*/

